DRUMS = "drums"
PIANO = "piano"
BRIGHT_PIANO = "bright_piano"
ELECTRIC_PIANO = "electric_piano"
HONKY_TONK = "honky_tonk"
ELECTRIC_PIANO_1 = "electric_piano_1"
ELECTRIC_PIANO_2 = "electric_piano_2"
HARPSICHORD = "harpsichord"
CLAVI = "clavi"
CELESTA = "celesta"
GLOCKENSPIEL = "glockenspiel"
MUSIC_BOX = "music_box"
VIBRAPHONE = "vibraphone"
MARIMBA = "marimba"
XYLOPHONE = "xylophone"
TUBULAR_BELLS = "tubular_bells"
DULCIMER = "dulcimer"
DRAWBAR_ORGAN = "drawbar_organ"
PERCUSSIVE_ORGAN = "percussive_organ"
ROCK_ORGAN = "rock_organ"
CHURCH_ORGAN = "church_organ"
REED_ORGAN = "reed_organ"
ACCORDION = "accordion"
HARMONICA = "harmonica"
TANGO_ACCORDION = "tango_accordion"
ACOUSTIC_GUITAR = "acoustic_guitar"
STEEL_GUITAR = "steel_guitar"
JAZZ_GUITAR = "jazz_guitar"
CLEAN_GUITAR = "clean_guitar"
MUTED_GUITAR = "muted_guitar"
OVERDRIVEN_GUITAR = "overdriven_guitar"
DISTORTION_GUITAR = "distortion_guitar"
HARMONIC_GUITAR = "harmonic_guitar"
ACOUSTIC_BASS = "acoustic_bass"
ELECTRIC_BASS_FINGER = "electric_bass_finger"
ELECTRIC_BASS_PICK = "electric_bass_pick"
FRETLESS_BASS = "fretless_bass"
SLAP_BASS_1 = "slap_bass_1"
SLAP_BASS_2 = "slap_bass_2"
SYNTH_BASS_1 = "synth_bass_1"
SYNTH_BASS_2 = "synth_bass_2"
VIOLIN = "violin"
VIOLA = "viola"
CELLO = "cello"
CONTRABASS = "contrabass"
TREMOLO_STRING = "tremolo_string"
PIZZICATO = "pizzicato"
HARP = "harp"
TIMPANI = "timpani"
STRING_ENSEMBLE_1 = "string_ensemble_1"
STRING_ENSEMBLE_2 = "string_ensemble_2"
SYNTH_STRING_1 = "synth_string_1"
SYNTH_STRING_2 = "synth_string_2"
CHOIR_AAHS = "choir_aahs"
CHOIR_OOHS = "choir_oohs"
SYNTH_CHOIR = "synth_choir"
ORCHESTRA_HIT = "orchestra_hit"
TRUMPET = "trumpet"
TROMBONE = "trombone"
TUBA = "tuba"
MUTED_TRUMPET = "muted_trumpet"
FRENCH_HORN = "french_horn"
BRASS_SECTION = "brass_section"
SYNTH_BRASS_1 = "synth_brass_1"
SYNTH_BRASS_2 = "synth_brass_2"
SOPRANO_SAX = "soprano_sax"
ALTO_SAX = "alto_sax"
TENOR_SAX = "tenor_sax"
BARITONE_SAX = "baritone_sax"
OBOE = "oboe"
ENGLISH_HORN = "english_horn"
BASSOON = "bassoon"
CLARINET = "clarinet"
PICCOLO = "piccolo"
FLUTE = "flute"
RECORDER = "recorder"
PAN_FLUTE = "pan_flute"
BLOWN_BOTTLE = "blown_bottle"
SHAKUHACHI = "shakuhachi"
WHISTLE = "whistle"
OCARINA = "ocarina"
SQUARE_LEAD = "square_lead"
SAWTOOTH_LEAD = "sawtooth_lead"
CALLIOPE_LEAD = "calliope_lead"
CHIFF_LEAD = "chiff_lead"
CHARANG_LEAD = "charang_lead"
VOICE_LEAD = "voice_lead"
FIFTHS_LEAD = "fifths_lead"
BASS_LEAD = "bass_lead"
PAD_NEW_AGE = "pad_new_age"
PAD_WARM = "pad_warm"
PAD_POLYSYNTH = "pad_polysynth"
PAD_CHOIR = "pad_choir"
PAD_BOWED = "pad_bowed"
PAD_METALLIC = "pad_metallic"
PAD_HALO = "pad_halo"
PAD_SWEEP = "pad_sweep"
FX_RAIN = "fx_rain"
FX_SOUNDTRACK = "fx_soundtrack"
FX_CRYSTAL = "fx_crystal"
FX_ATMOSPHERE = "fx_atmosphere"
FX_BRIGHTNESS = "fx_brightness"
FX_GLOBLINS = "fx_globlins"
FX_ECHOES = "fx_echoes"
FX_SCI_FI = "fx_sci_fi"
SITAR = "sitar"
BANJO = "banjo"
SHAMISEN = "shamisen"
KOTO = "koto"
KALIMBA = "kalimba"
BAGPIPE = "bagpipe"
FIDDLE = "fiddle"
SHANAI = "shanai"
TINKLE_BELL = "tinkle_bell"
AGOGO = "agogo"
STEEL_DRUMS = "steel_drums"
WOODBLOCK = "woodblock"
TAIKO_DRUM = "taiko_drum"
MELODIC_TOM = "melodic_tom"
SYNTH_DRUM = "synth_drum"
REVERSE_CYMBAL = "reverse_cymbal"
GUITAR_FRET_NOISE = "guitar_fret_noise"
BREATH_NOISE = "breath_noise"
SEASHORE = "seashore"
BIRD_TWEET = "bird_tweet"
TELEPHONE_RING = "telephone_ring"
HELICOPTER = "helicopter"
APPLAUSE = "applause"
GUNSHOT = "gunshot"

ALL_INSTRUMENTS = [
    DRUMS,
    PIANO,
    BRIGHT_PIANO,
    ELECTRIC_PIANO,
    HONKY_TONK,
    ELECTRIC_PIANO_1,
    ELECTRIC_PIANO_2,
    HARPSICHORD,
    CLAVI,
    CELESTA,
    GLOCKENSPIEL,
    MUSIC_BOX,
    VIBRAPHONE,
    MARIMBA,
    XYLOPHONE,
    TUBULAR_BELLS,
    DULCIMER,
    DRAWBAR_ORGAN,
    PERCUSSIVE_ORGAN,
    ROCK_ORGAN,
    CHURCH_ORGAN,
    REED_ORGAN,
    ACCORDION,
    HARMONICA,
    TANGO_ACCORDION,
    ACOUSTIC_GUITAR,
    STEEL_GUITAR,
    JAZZ_GUITAR,
    CLEAN_GUITAR,
    MUTED_GUITAR,
    OVERDRIVEN_GUITAR,
    DISTORTION_GUITAR,
    HARMONIC_GUITAR,
    ACOUSTIC_BASS,
    ELECTRIC_BASS_FINGER,
    ELECTRIC_BASS_PICK,
    FRETLESS_BASS,
    SLAP_BASS_1,
    SLAP_BASS_2,
    SYNTH_BASS_1,
    SYNTH_BASS_2,
    VIOLIN,
    VIOLA,
    CELLO,
    CONTRABASS,
    TREMOLO_STRING,
    PIZZICATO,
    HARP,
    TIMPANI,
    STRING_ENSEMBLE_1,
    STRING_ENSEMBLE_2,
    SYNTH_STRING_1,
    SYNTH_STRING_2,
    CHOIR_AAHS,
    CHOIR_OOHS,
    SYNTH_CHOIR,
    ORCHESTRA_HIT,
    TRUMPET,
    TROMBONE,
    TUBA,
    MUTED_TRUMPET,
    FRENCH_HORN,
    BRASS_SECTION,
    SYNTH_BRASS_1,
    SYNTH_BRASS_2,
    SOPRANO_SAX,
    ALTO_SAX,
    TENOR_SAX,
    BARITONE_SAX,
    OBOE,
    ENGLISH_HORN,
    BASSOON,
    CLARINET,
    PICCOLO,
    FLUTE,
    RECORDER,
    PAN_FLUTE,
    BLOWN_BOTTLE,
    SHAKUHACHI,
    WHISTLE,
    OCARINA,
    SQUARE_LEAD,
    SAWTOOTH_LEAD,
    CALLIOPE_LEAD,
    CHIFF_LEAD,
    CHARANG_LEAD,
    VOICE_LEAD,
    FIFTHS_LEAD,
    BASS_LEAD,
    PAD_NEW_AGE,
    PAD_WARM,
    PAD_POLYSYNTH,
    PAD_CHOIR,
    PAD_BOWED,
    PAD_METALLIC,
    PAD_HALO,
    PAD_SWEEP,
    FX_RAIN,
    FX_SOUNDTRACK,
    FX_CRYSTAL,
    FX_ATMOSPHERE,
    FX_BRIGHTNESS,
    FX_GLOBLINS,
    FX_ECHOES,
    FX_SCI_FI,
    SITAR,
    BANJO,
    SHAMISEN,
    KOTO,
    KALIMBA,
    BAGPIPE,
    FIDDLE,
    SHANAI,
    TINKLE_BELL,
    AGOGO,
    STEEL_DRUMS,
    WOODBLOCK,
    TAIKO_DRUM,
    MELODIC_TOM,
    SYNTH_DRUM,
    REVERSE_CYMBAL,
    GUITAR_FRET_NOISE,
    BREATH_NOISE,
    SEASHORE,
    BIRD_TWEET,
    TELEPHONE_RING,
    HELICOPTER,
    APPLAUSE,
    GUNSHOT
]

