
from maidi.analysis.control_features.density import DensityTagsProvider
from maidi.analysis.control_features.polyphony import MinMaxPolyphonyTagsProvider
from maidi.analysis.control_features.register import MinMaxRegisterTagsProvider
from maidi.analysis.control_features.special_notes import SpecialNotesTagsProvider





__all__ = [
    "DensityTagsProvider",
    "MinMaxPolyphonyTagsProvider",
    "MinMaxRegisterTagsProvider",
    "SpecialNotesTagsProvider",
]
