from .parser import Parser
from .constants import *
from .score import MidiScore
from .analysis import TagsProvider, ScoreTagger, TagManager, Tags
from .chords import ChordManager
from .integrations import MusicLangAPI

__version__ = "0.15.1"