MODEL_CONTROL_MASKING_LARGE = "control_masking_large"
MODEL_MASKING_LARGE = "masking_large"
MODEL_MASKING_SMALL = "masking_small"
MODEL_CONTROL_MASKING_OPTIMIZED = "control_masking_optimized"

MODELS = [MODEL_CONTROL_MASKING_LARGE, MODEL_MASKING_LARGE, MODEL_MASKING_SMALL, MODEL_CONTROL_MASKING_OPTIMIZED]
ONLY_SYNC_MODELS = []
